# The contents of this file are subject to the Common Public Attribution License
# Version 1.0 (the “License”); you may not use this file except in compliance
# with the License. You may obtain a copy of the License at
# https://myou.dev/licenses/LICENSE-CPAL. The License is based on the Mozilla
# Public License Version 1.1 but Sections 14 and 15 have been added to cover use
# of software over a computer network and provide for limited attribution for
# the Original Developer. In addition, Exhibit A has been modified to be
# consistent with Exhibit B.
#
# Software distributed under the License is distributed on an “AS IS” basis,
# WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
# the specific language governing rights and limitations under the License.
#
# The Original Code is Myou Engine.
#
# the Original Developer is the Initial Developer.
#
# The Initial Developer of the Original Code is the Myou Engine developers.
# All portions of the code written by the Myou Engine developers are Copyright
# (c) 2024. All Rights Reserved.
#
# Alternatively, the contents of this file may be used under the terms of the
# GNU Affero General Public License version 3 (the [AGPL-3] License), in which
# case the provisions of [AGPL-3] License are applicable instead of those above.
#
# If you wish to allow use of your version of this file only under the terms of
# the [AGPL-3] License and not to allow others to use your version of this file
# under the CPAL, indicate your decision by deleting the provisions above and
# replace them with the notice and other provisions required by the [AGPL-3]
# License. If you do not delete the provisions above, a recipient may use your
# version of this file under either the CPAL or the [AGPL-3] License.

import std/strutils
import ../types
import vmath except Quat
import ../graphics/framebuffer
import ../graphics/material
import ../graphics/render
import ../graphics/texture
import ../graphics/ubo
# import ../objects/light
import ../objects/mesh
import ../quat
import ../scene
import ../util

# TODO: Move these base methods
method renderShadow*(self: ShadowManager, camera: Camera) {.base.} =
    discard
method destroy*(self: ShadowManager) {.base.} =
    discard

when defined(nimdoc):
    # TODO: ShadowMapUniform goes elsewhere
    type TYPES* = SimpleShadowManager | ShadowMapUniform

# Z depth is not modified, only X and Y for UVs
const LIGHT_PROJ_TO_DEPTH = mat4(0.5, 0, 0, 0,  0, 0.5, 0, 0, 
                                 0, 0, 1, 0,  0.5, 0.5, 0, 1)

when not defined(release):
    var debug_mesh: Mesh

proc newSimpleShadowManager*(light: Light; fb_size: int;
                             near, far: float32 = 0.0;
                             bias: float32 = 0.001;
                            ): SimpleShadowManager =
    
    new(result)
    result.engine = light.engine
    result.light = light
    result.depth_range = (near, far)
    result.bias = bias
    result.auto_render = true
    light.shadows.add result
    let self = result
    result.engine.renderer.enqueue proc()=
        # TODO: allow configuring use of a depth-only pass with this code
        # and with sampler2DShadow support in texture.nim

        # self.material = light.engine.newMaterial("simple_shadow", nil,
        #     fragment="",
        # )
        # self.framebuffer = light.engine.newFramebuffer(fb_size, fb_size,
        #     RGB_u8, DepthTexture, depth_format = Depth_u16, filter=Linear)
        # self.texture = self.framebuffer.depth_texture

        self.material = light.engine.newMaterial("simple_shadow", nil,
            fragment=dedent """
            in vec4 ppos;
            out float color;
            void main(){
                color = ppos.z/ppos.w;
            }""",
            varyings = @[Varying(vtype: ProjPosition, varname: "ppos")]
        )
        self.framebuffer = light.engine.newFramebuffer(fb_size, fb_size,
            R_f16, DepthRenderBuffer, filter=Linear)
        self.texture = self.framebuffer.texture
        self.texture.setExtrapolation Clamp
        when not defined(release):
            debug_mesh = self.engine.newMesh(vertex_count = 100, draw_method = Lines)
            debug_mesh.materials.add self.engine.newSolidMaterial("white", vec4(0,1,0,1))
            debug_mesh.visible = false
            # TODO: why the FUCK do I need this here?
            self.engine.renderer.enqueue proc()=
                light.scene.add_object debug_mesh

method destroy*(self: SimpleShadowManager) =
    if self.material != nil:
        self.material.destroy()
    if self.framebuffer != nil:
        self.framebuffer.destroy()

when not defined(release):
    proc show_mat(mat: Mat4) =
        var points = newSeqOfCap[Vec3](8)
        for z in [-1'f32,1'f32]:
            for y in [-1'f32,1'f32]:
                for x in [-1'f32,1'f32]:
                    let v = mat * vec4(x,y,z,1)
                    points.add v.xyz/v.w
        for i in [0,1,1,3,3,2,2,0,4,5,5,7,7,6,6,4,0,4,1,5,2,6,3,7]:
            debug_mesh.add_vertex points[i], vec4(1)
        debug_mesh.data.update_varray
        debug_mesh.visible = true

method renderShadow*(self: SimpleShadowManager, scene: Scene, bounding_points: seq[Vec3], min_z: float32, index: int)

method renderShadow*(self: SimpleShadowManager, camera: Camera, index: int) =
    # first we'll make a list of points that matches the corners of the camera
    # frustum and given depth range

    # it assumes that all world matrices are already up to date
    # TODO: accept a cameradata instead?

    when not defined(release):
        debug_mesh.clear_vertices()
    
    template depth_to_screen(d: float32, proj: Mat4): float32 =
        # TODO: make more efficient and put in Camera
        let v = proj * vec4(0,0,-d,1)
        clamp(v.z/v.w,-1.0, 1.0)

    # get frustum points, in world space
    let screen_depth = if self.depth_range == (0'f32,0'f32):
        [-1'f32, 1'f32]
    else: [
        depth_to_screen(self.depth_range[0], camera.projection_matrix),
        depth_to_screen(self.depth_range[1], camera.projection_matrix),
    ]
    let mat = camera.world_matrix.remove_scale_skew * camera.projection_matrix_inverse
    # show_mat mat
    var points = newSeqOfCap[Vec3](8)
    for z in screen_depth:
        for y in [-1'f32,1'f32]:
            for x in [-1'f32,1'f32]:
                let v = mat * vec4(x,y,z,1)
                points.add v.xyz/v.w
    # for i in [0,1,1,3,3,2,2,0,4,5,5,7,7,6,6,4,0,4,1,5,2,6,3,7]:
    #     debug_mesh.add_vertex points[i], vec4(1)

    var zrange = self.depth_range[1] - self.depth_range[0]
    if zrange == 0.0:
        zrange = camera.far_plane - camera.near_plane
    # extend z range an arbitrary amount
    # TODO: make it configurable
    # or better: clamp polygons in the shader!
    self.renderShadow(camera.scene, points, min_z=zrange*4, index)

method renderShadow*(self: SimpleShadowManager, scene: Scene, bounding_points: seq[Vec3], min_z: float32, index: int) =
    # Find the projection matrix that fits all bounding points
    # TODO: find bounds of bounding sphere instead
    # TODO: option to favor vertical lines?
    when not defined(release):
        debug_mesh.clear_vertices()
    let rotm = self.light.world_matrix.to_mat3_rotation
    let rotmi = rotm.inverse
    var pmin = vec3(Inf)
    var pmax = vec3(-Inf)
    for p in bounding_points:
        let p = rotmi * p
        pmin = min(pmin, p)
        pmax = max(pmax, p)
    # TODO: extend pmin.z so it never falls inside the complete frustum
    pmax.z = max(pmin.z+min_z, pmax.z)
    let mid = mix(pmin, pmax, 0.5)
    var scale = (pmax-pmin) * vec3(0.5, 0.5, -0.5)
    scale.xy = vec2(max(scale.x, scale.y))
    let proj = scale(1'f32/scale)
    let light_matrix = rotm.to_mat4 * translate(mid)
    
    # show_mat inverse(proj * inverse(light_matrix))


    self.texture.unbind()
    self.framebuffer.enable()
    self.framebuffer.clear(color=vec4(1))
    let cd = newRenderCameraData(light_matrix,
        proj, proj.get_culling_planes(), vec2(self.framebuffer.width.float32))
    
    let rm = self.engine.renderer
    if not rm.initialized: # TODO: enqueue but only once
        return
    let material = self.material
    for ob in scene.mesh_passes[0]:
        if ob.visible:
            rm.draw_mesh(ob, ob.world_matrix, cd, 0, material)

    self.framebuffer.disable()

    scene.shadow_maps[index] = self.texture
    var depth_matrix = LIGHT_PROJ_TO_DEPTH * proj * inverse(light_matrix)
    # vmath doesn't let me do this properly
    depth_matrix[3,2] = depth_matrix[3,2] - self.bias
    # cast[ptr float32](depth_matrix[3,2].addr)[] -= self.bias
    scene.shadow_maps_ubo.storage(ShadowMapUniform)[index] =
        ShadowMapUniform(depth_matrix: depth_matrix,
            tex_size: self.framebuffer.width.float32)
    
    # TODO: do this outside
    scene.shadow_maps_ubo.update()

